@@ -1,19 +1,21 @@
-local lfs = require('lfs')
-local helpers = require('test.functional.helpers')(after_each)
+local t = require('test.testutil')
+local n = require('test.functional.testnvim')()
 local Screen = require('test.functional.ui.screen')
 
-local clear = helpers.clear
-local command = helpers.command
-local get_pathsep = helpers.get_pathsep
-local iswin = helpers.iswin
-local eq = helpers.eq
-local neq = helpers.neq
-local funcs = helpers.funcs
-local matches = helpers.matches
-local pesc = helpers.pesc
-local rmdir = helpers.rmdir
-local sleep = helpers.sleep
-local meths = helpers.meths
+local clear = n.clear
+local command = n.command
+local get_pathsep = n.get_pathsep
+local eq = t.eq
+local neq = t.neq
+local fn = n.fn
+local matches = t.matches
+local pesc = vim.pesc
+local rmdir = n.rmdir
+local sleep = vim.uv.sleep
+local api = n.api
+local skip = t.skip
+local is_os = t.is_os
+local mkdir = t.mkdir
 
 local file_prefix = 'Xtest-functional-ex_cmds-mksession_spec'
 
@@ -23,7 +25,7 @@ describe(':mksession', function()
 
   before_each(function()
     clear()
-    lfs.mkdir(tab_dir)
+    mkdir(tab_dir)
   end)
 
   after_each(function()
@@ -49,8 +51,8 @@ describe(':mksession', function()
     -- Restore session.
     command('source ' .. session_file)
 
-    eq(funcs.winbufnr(1), funcs.winbufnr(2))
-    neq(funcs.winbufnr(1), funcs.winbufnr(3))
+    eq(fn.winbufnr(1), fn.winbufnr(2))
+    neq(fn.winbufnr(1), fn.winbufnr(3))
   end)
 
   -- common testing procedure for testing "sessionoptions-=terminal"
@@ -65,64 +67,61 @@ describe(':mksession', function()
     -- Restore session.
     command('source ' .. session_file)
 
-    eq(expected_buf_count, #meths.list_bufs())
+    eq(expected_buf_count, #api.nvim_list_bufs())
   end
 
-  it(
-    'do not restore :terminal if not set in sessionoptions, terminal in current window #13078',
-    function()
-      local tmpfile_base = file_prefix .. '-tmpfile'
-      command('edit ' .. tmpfile_base)
-      command('terminal')
+  it('do not restore :terminal if not set in sessionoptions, terminal in curwin #13078', function()
+    local tmpfile_base = file_prefix .. '-tmpfile'
+    command('edit ' .. tmpfile_base)
+    command('terminal')
 
-      local buf_count = #meths.list_bufs()
-      eq(2, buf_count)
+    local buf_count = #api.nvim_list_bufs()
+    eq(2, buf_count)
 
-      eq('terminal', meths.buf_get_option(0, 'buftype'))
+    eq('terminal', api.nvim_get_option_value('buftype', {}))
 
-      test_terminal_session_disabled(2)
+    test_terminal_session_disabled(2)
 
-      -- no terminal should be set. As a side effect we end up with a blank buffer
-      eq('', meths.buf_get_option(meths.list_bufs()[1], 'buftype'))
-      eq('', meths.buf_get_option(meths.list_bufs()[2], 'buftype'))
-    end
-  )
+    -- no terminal should be set. As a side effect we end up with a blank buffer
+    eq('', api.nvim_get_option_value('buftype', { buf = api.nvim_list_bufs()[1] }))
+    eq('', api.nvim_get_option_value('buftype', { buf = api.nvim_list_bufs()[2] }))
+  end)
 
   it('do not restore :terminal if not set in sessionoptions, terminal hidden #13078', function()
     command('terminal')
-    local terminal_bufnr = meths.get_current_buf()
+    local terminal_bufnr = api.nvim_get_current_buf()
 
     local tmpfile_base = file_prefix .. '-tmpfile'
     -- make terminal hidden by opening a new file
     command('edit ' .. tmpfile_base .. '1')
 
-    local buf_count = #meths.list_bufs()
+    local buf_count = #api.nvim_list_bufs()
     eq(2, buf_count)
 
-    eq(1, funcs.getbufinfo(terminal_bufnr)[1].hidden)
+    eq(1, fn.getbufinfo(terminal_bufnr)[1].hidden)
 
     test_terminal_session_disabled(1)
 
     -- no terminal should exist here
-    neq('', meths.buf_get_name(meths.list_bufs()[1]))
+    neq('', api.nvim_buf_get_name(api.nvim_list_bufs()[1]))
   end)
 
   it('do not restore :terminal if not set in sessionoptions, only buffer #13078', function()
     command('terminal')
-    eq('terminal', meths.buf_get_option(0, 'buftype'))
+    eq('terminal', api.nvim_get_option_value('buftype', {}))
 
-    local buf_count = #meths.list_bufs()
+    local buf_count = #api.nvim_list_bufs()
     eq(1, buf_count)
 
     test_terminal_session_disabled(1)
 
     -- no terminal should be set
-    eq('', meths.buf_get_option(0, 'buftype'))
+    eq('', api.nvim_get_option_value('buftype', {}))
   end)
 
   it('restores tab-local working directories', function()
     local tmpfile_base = file_prefix .. '-tmpfile'
-    local cwd_dir = funcs.getcwd()
+    local cwd_dir = fn.getcwd()
 
     -- :mksession does not save empty tabs, so create some buffers.
     command('edit ' .. tmpfile_base .. '1')
@@ -138,15 +137,15 @@ describe(':mksession', function()
     command('source ' .. session_file)
     -- First tab should have the original working directory.
     command('tabnext 1')
-    eq(cwd_dir, funcs.getcwd())
+    eq(cwd_dir, fn.getcwd())
     -- Second tab should have the tab-local working directory.
     command('tabnext 2')
-    eq(cwd_dir .. get_pathsep() .. tab_dir, funcs.getcwd())
+    eq(cwd_dir .. get_pathsep() .. tab_dir, fn.getcwd())
   end)
 
   it('restores buffers with tab-local CWD', function()
     local tmpfile_base = file_prefix .. '-tmpfile'
-    local cwd_dir = funcs.getcwd()
+    local cwd_dir = fn.getcwd()
     local session_path = cwd_dir .. get_pathsep() .. session_file
 
     command('edit ' .. tmpfile_base .. '1')
@@ -162,13 +161,15 @@ describe(':mksession', function()
     -- Use :silent to avoid press-enter prompt due to long path
     command('silent source ' .. session_path)
     command('tabnext 1')
-    eq(cwd_dir .. get_pathsep() .. tmpfile_base .. '1', funcs.expand('%:p'))
+    eq(cwd_dir .. get_pathsep() .. tmpfile_base .. '1', fn.expand('%:p'))
     command('tabnext 2')
-    eq(cwd_dir .. get_pathsep() .. tmpfile_base .. '2', funcs.expand('%:p'))
+    eq(cwd_dir .. get_pathsep() .. tmpfile_base .. '2', fn.expand('%:p'))
   end)
 
   it('restores CWD for :terminal buffers #11288', function()
-    local cwd_dir = funcs.fnamemodify('.', ':p:~'):gsub([[[\/]*$]], '')
+    skip(is_os('win'), 'causes rmdir() to fail')
+
+    local cwd_dir = fn.fnamemodify('.', ':p:~'):gsub([[[\/]*$]], '')
     cwd_dir = cwd_dir:gsub([[\]], '/') -- :mksession always uses unix slashes.
     local session_path = cwd_dir .. '/' .. session_file
 
@@ -177,7 +178,7 @@ describe(':mksession', function()
     command('cd ' .. cwd_dir)
     command('mksession ' .. session_path)
     command('%bwipeout!')
-    if iswin() then
+    if is_os('win') then
       sleep(100) -- Make sure all child processes have exited.
     end
 
@@ -186,21 +187,18 @@ describe(':mksession', function()
     command('silent source ' .. session_path)
 
     local expected_cwd = cwd_dir .. '/' .. tab_dir
-    matches('^term://' .. pesc(expected_cwd) .. '//%d+:', funcs.expand('%'))
+    matches('^term://' .. pesc(expected_cwd) .. '//%d+:', fn.expand('%'))
     command('%bwipeout!')
-    if iswin() then
+    if is_os('win') then
       sleep(100) -- Make sure all child processes have exited.
     end
   end)
 
   it('restores CWD for :terminal buffer at root directory #16988', function()
-    if iswin() then
-      pending('N/A for Windows')
-      return
-    end
+    skip(is_os('win'), 'N/A for Windows')
 
     local screen
-    local cwd_dir = funcs.fnamemodify('.', ':p:~'):gsub([[[\/]*$]], '')
+    local cwd_dir = fn.fnamemodify('.', ':p:~'):gsub([[[\/]*$]], '')
     local session_path = cwd_dir .. '/' .. session_file
 
     screen = Screen.new(50, 6)
@@ -209,9 +207,7 @@ describe(':mksession', function()
       ^/                                                 |
                                                         |
       [Process exited 0]                                |
-                                                        |
-                                                        |
-                                                        |
+                                                        |*3
     ]]
 
     command('cd /')
@@ -238,7 +234,7 @@ describe(':mksession', function()
     local tmpfile = file_prefix .. '-tmpfile-float'
 
     command('edit ' .. tmpfile)
-    local buf = meths.create_buf(false, true)
+    local buf = api.nvim_create_buf(false, true)
     local config = {
       relative = 'editor',
       focusable = false,
@@ -248,8 +244,8 @@ describe(':mksession', function()
       col = 1,
       style = 'minimal',
     }
-    meths.open_win(buf, false, config)
-    local cmdheight = meths.get_option('cmdheight')
+    api.nvim_open_win(buf, false, config)
+    local cmdheight = api.nvim_get_option_value('cmdheight', {})
     command('mksession ' .. session_file)
 
     -- Create a new test instance of Nvim.
@@ -257,12 +253,12 @@ describe(':mksession', function()
 
     command('source ' .. session_file)
 
-    eq(tmpfile, funcs.expand('%'))
+    eq(tmpfile, fn.expand('%'))
     -- Check that there is only a single window, which indicates the floating
     -- window was not restored.
-    eq(1, funcs.winnr('$'))
+    eq(1, fn.winnr('$'))
     -- The command-line height should remain the same as it was.
-    eq(cmdheight, meths.get_option('cmdheight'))
+    eq(cmdheight, api.nvim_get_option_value('cmdheight', {}))
 
     os.remove(tmpfile)
   end)
